(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Ceta_Checker
imports Ceta
begin

subsection {* User exceptions in generated code *}
(* FIXME:  will be moved to Main. update: did not happen for Isabelle2012. *)

definition error :: "String.literal \<Rightarrow> 'a \<Rightarrow> 'a" 
where [code del]: "error msg x = x"

code_const error
  (Eval "(fn msg =>/ fn x =>/ error msg)")

code_reserved Eval error

subsection {* Checker *}

text {* Assemble a checker with a simple interface from the bits and pieces provided by IsaFoR. *}

(* FIXME: Shouldn't the errors be uniform (string vs. show)??? *)
fun release_shows_error :: "'a \<Rightarrow> shows + 'a \<Rightarrow> 'a"
where
  "release_shows_error d (Inl msg) = error (implode (msg [])) d"
| "release_shows_error d (Inr x) = x"

fun release_string_error :: "'a \<Rightarrow> 'a sparse_result \<Rightarrow> 'a"
where
  "release_string_error d (PError msg) = error (implode msg) d"
| "release_string_error d (PResult x) = x"

lemma [code]: "release_string_error d (parseresult (Inl msg)) = error (implode msg) d" by simp
lemma [code]: "release_string_error d (parseresult (Inr x)) = x" by simp


lemma inj_map_term_NF_terms: assumes inj: "inj fg"
  and NF: "t \<in> NF_terms Q"
  shows "map_term fg t \<in> NF_terms (map_term fg ` Q)"
proof -
  let ?m = "map_term fg"
  let ?Q = "?m ` Q"
  let ?gf = "the_inv fg"
  have id: "?gf o fg = id"
    by (rule ext, insert the_inv_f_f[OF inj], auto)
  {
    fix s
    have "(?m t, s) \<notin> rstep (Id_on ?Q)" 
    proof
      assume "(?m t, s) \<in> rstep (Id_on ?Q)"
      from rstep_imp_map_rstep[OF this, of ?gf, unfolded map_term_comp id map_term_ident]
      have "t \<notin> NF (rstep (map_trs ?gf (Id_on ?Q)))" by auto
      also have "map_trs ?gf (Id_on ?Q) = Id_on Q"
        unfolding map_trs.simps Id_on_def using the_inv_f_f[OF inj] 
        by (force simp: id map_term_comp)
      finally show False using NF ..
    qed
  }
  thus ?thesis by auto
qed

lemma inj_map_term_qrstep: assumes inj: "inj fg"
  and step: "(s,t) \<in> qrstep nfs Q R"
  shows "(map_term fg s, map_term fg t) \<in> qrstep nfs (map_term fg ` Q) (map_trs fg R)" 
proof - 
  let ?m = "map_term fg"
  let ?Q = "?m ` Q"
  let ?R = "map_trs fg R"
  from step obtain C l r \<sigma> where lr: "(l,r) \<in> R" and s: "s = C \<langle> l \<cdot> \<sigma> \<rangle>" and t: "t = C \<langle> r \<cdot> \<sigma> \<rangle>"
  and NF: "\<forall> u \<lhd> l \<cdot> \<sigma>. u \<in> NF_terms Q" 
  and nfs: "NF_subst nfs (l,r) \<sigma> Q" by auto
  note args_conv = NF_terms_args_conv
  note distrib = map_ctxt_distrib map_subst_distrib
  from lr have lr: "(?m l, ?m r) \<in> ?R" unfolding map_trs.simps by force  
  show ?thesis unfolding s t distrib
  proof (rule qrstepI[OF _ lr refl refl], unfold distrib[symmetric] args_conv[symmetric], rule)
    fix u
    assume u: "u \<in> set (args (?m (l \<cdot> \<sigma>)))"
    then obtain f ls where ls: "l \<cdot> \<sigma> = Fun f ls" by (cases "l \<cdot> \<sigma>", auto)
    from u[unfolded ls] obtain li where li: "li \<in> set ls" and u: "u = ?m li" by auto
    from li ls have "li \<in> set (args (l \<cdot> \<sigma>))" by auto
    with NF[unfolded args_conv[symmetric]] have li: "li \<in> NF_terms Q" by auto
    from inj_map_term_NF_terms[OF inj li]
    show "u \<in> NF_terms ?Q" unfolding u .
  next
    show "NF_subst nfs (?m l, ?m r) (map_subst fg \<sigma>) ?Q"
    proof
      fix x 
      assume nfs and "x \<in> vars_term (?m l) \<or> x \<in> vars_term (?m r)"
      hence "x \<in> vars_rule (l,r)" unfolding vars_rule_def by auto
      with nfs[unfolded NF_subst_def] `nfs` have "\<sigma> x \<in> NF_terms Q" by auto
      from inj_map_term_NF_terms[OF inj this]
      show "map_subst fg \<sigma> x \<in> NF_terms ?Q" by simp
    qed
  qed
qed

lemma SN_map: assumes SN: "SN A"
  and map: "\<And> x y. (x,y) \<in> B \<Longrightarrow> (f x, f y) \<in> A"
  shows "SN B"
proof 
  fix g
  assume steps: "\<forall> i. (g i, g (Suc i)) \<in> B"
  obtain h where h: "h \<equiv> \<lambda> i. f (g i)" by auto
  with map steps have "\<forall> i. (h i, h (Suc i)) \<in> A" by auto
  with SN show False by auto
qed

lemma inj_map_term_qrstep_SN: assumes inj: "inj fg"
  and SN: "SN (qrstep nfs (map_term fg ` Q) (map_trs fg R))"
  shows "SN (qrstep nfs Q R)"
  by (rule SN_map[OF SN inj_map_term_qrstep[OF inj]])

definition ceta_checker :: "(string, string) rules \<Rightarrow> (string,string)term list \<Rightarrow>
  (string, label_type, string) trs_proof \<Rightarrow> bool"
where
  "ceta_checker trs str prf = 
    release_shows_error False 
      (check_trs_proof False
        (shows_string ''1'')
        tp_impl
        dpp_impl
        (tp_ops.mk tp_impl True (map init_lab_term str) (init_lab_rules trs) [])
        prf
      \<guillemotright>= (\<lambda>x. Error_Monad.return True))"

lemma ceta_checker:
  assumes check: "ceta_checker trs str prf"
  shows "SN (qrstep True (set str) (set trs))"
proof -
  let ?Q = "map init_lab_term str ::((string, label_type) lab, string) term list"
  let ?S = "[]::((string, label_type) lab, string) rules"
  let ?QQ = "init_lab_term ` (set str)"
  let ?R = "init_lab_rules trs::((string, label_type) lab, string) rules"
  let ?RR = "init_lab_trs (set trs)::((string, label_type) lab, string) trs"
  let ?call = "check_trs_proof False
    (shows_string ''1'') tp_impl dpp_impl (tp_ops.mk tp_impl True ?Q ?R ?S) prf"
  show ?thesis 
  proof (cases ?call)
    case Inl with check show ?thesis by (auto simp: error_def ceta_checker_def)
  next
    case Inr
    then have "isOK (?call)" by simp
    from check_trs_proof_sound[OF tp_impl dpp_impl this]
    have "SN (qrstep True (set ?Q) (set ?R))"
      unfolding tp_spec.mk_sound[OF tp_impl] SN_qrel_def split
      by simp
    hence SN: "SN (qrstep True ?QQ ?RR)"
      unfolding map_trs.simps by auto
    show ?thesis 
    proof (rule inj_map_term_qrstep_SN[OF _ SN])
      show "inj init_lab" unfolding inj_on_def by simp
    qed
  qed
qed

definition dummy_proof :: "(string, label_type, string) trs_proof" where
  "dummy_proof = RisEmpty"

definition parse_trs_proof :: "string \<Rightarrow> Code_Evaluation.term" where
  "parse_trs_proof s = Code_Evaluation.term_of 
  (case release_string_error
    (TRS_Proof No_Strategy [] None dummy_proof) (parse_cert_problem s) of
    TRS_Proof _ _ _ prf \<Rightarrow> prf
  | _ \<Rightarrow> error (STR ''not a termination proof'') dummy_proof)"

ML {*
structure Ceta_Parser =
struct
  val read_cpf = 
    String.explode
    #> @{code parse_trs_proof}
    #> Pattern.rewrite_term @{theory} 
      [(term_of @{cpat "Frct (?x, ?y)"}, term_of @{cpat "Fract ?x ?y"})] []
end
*}

(* if the following is enabled, then ceta_simp and ceta_checker_simp do not work any longer *)
code_reflect Ceta_Checker
  datatypes trs_proof = "_" 
  and dp_proof = "_" and redtriple_impl = "_" and root_redtriple_impl = "_"
  and projL = "_" and sl_inter = "_" and arithFun = "_" and redtriple_impl = "_"
  and af_entry = "_" and arctic_rat = "_" and arctic = "_" and tpoly = "_"
  and lab = "_" and nibble = "_" and sum = "_" and pos = "_"
  and gctxt = "_" and "term" = "_" and order_tag = "_"
  and scg = "_" and rat = "_" and nat = "_" and join_info = "_"
   (*FIXME:other types must also be mentioned here and this should correspond somehow to 
           the list mentioned in termfun.ML *)
  functions ceta_checker Fract Trueprop
(* remove code equation (already installed anyway), to avoid expensive collection of all other code equations *)
declare ceta_checker_def[code del]

end
