(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Dependency_Graph_Impl
imports
  Graph
  Edg_Impl
  QDP_Framework
  Icap_Impl
  Dependency_Graph
  Innermost_Usable_Rules_Impl
begin

definition
  is_iedg_edge_dpp ::
    "('d,'f :: {show,linorder},string)dpp_ops \<Rightarrow> 'd \<Rightarrow> ('f,string)rule \<Rightarrow> ('f,string)term \<Rightarrow> bool"
where
  "is_iedg_edge_dpp I d \<equiv> let qnf = dpp_ops.is_QNF I d;
                              ic = icap_impl_dpp_mv I d;
                              R = dpp_ops.rules I d;
                              urules = inn_usable_rules_pair I d in 
      (\<lambda> (s,t). let cst = ic [s] t;
                    urls = reverse_rules (urules (s,t));
                    ic'  = icap_impl' urls []
         in (\<lambda> u.
            (case mgu_class cst u of None \<Rightarrow> False | Some \<mu>
                \<Rightarrow> qnf (mv_xvar s \<cdot> \<mu>) \<and> qnf (mv_yvar u \<cdot> \<mu>))
           \<and> (let cu = ic' [] u in 
               case mgu_class cu t of None \<Rightarrow> False | Some \<mu>
                \<Rightarrow> qnf (mv_yvar s \<cdot> \<mu>) 
             )))"

lemma is_iedg_edge_dpp_sound:
  fixes I::"('d, 'f::{show,linorder}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and r: "r \<equiv> dpp_ops.rules I d"
      and nfs: "nfs \<equiv> dpp_ops.nfs I d"
      and m: "m \<equiv> dpp_ops.minimal I d"
  assumes I: "dpp_spec I"
  and sNF: "s \<cdot> \<sigma> \<in> NF_terms (set q)"
  and uNF: "u \<cdot> \<tau> \<in> NF_terms (set q)"
  and nfs1: "NF_subst nfs (s,t) \<sigma> (set q)"
  and nfs2: "NF_subst nfs (u,v) \<tau> (set q)"
  and tSN: "m \<Longrightarrow> SN_on (qrstep nfs (set q) (set r)) {t \<cdot> \<sigma>}" 
  and steps: "(t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep nfs (set q) (set r))^*"
  shows "is_iedg_edge_dpp I d (s,t) u"
proof -
  interpret dpp_spec I by fact
  let ?U = "inn_usable_rules_pair I d (s,t)"
  let ?nfs = "NFS d"
  let ?m = "M d"
  obtain U where U: "U = reverse_rules ?U" by auto
  from sNF uNF nfs1 nfs2 steps tSN have "((s,t),(u,v)) \<in> DG ?nfs ?m UNIV (set (dpp_ops.Q I d)) (set (dpp_ops.rules I d))" unfolding 
    DG_def r q nfs m by auto  
  with IEDG[of ?nfs ?m] have edge: "((s,t),(u,v)) \<in> IEDG UNIV (set (dpp_ops.Q I d)) (set (dpp_ops.rules I d))" by auto
  from inn_usable_rules_pair[OF I refl steps[unfolded q r nfs] sNF[unfolded q nfs] nfs1[unfolded q nfs] tSN[unfolded q r nfs m]]
  have "(t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep nfs (set q) (set ?U))^*" unfolding q nfs .
  hence "(u \<cdot> \<tau>, t \<cdot> \<sigma>) \<in> (rstep (set U))^*"
    unfolding U reverse_rules 
    unfolding reverseTrs[symmetric]
    unfolding rtrancl_converse
    using rtrancl_mono[OF qrstep_subset_rstep] by auto
  hence steps: "(u \<cdot> \<tau>, t \<cdot> \<sigma>) \<in> (qrstep nfs {} (set U))^*" by simp
  from icap_mv_mgu[OF steps, of "{}"] obtain \<mu> \<delta> where mgu: 
    "mgu_class (icap_mv (set U) {} {} u) t = Some \<mu>"
    and \<sigma>: "\<And> s. s \<cdot> \<sigma> = mv_yvar s \<cdot> \<mu> \<cdot> \<delta>" by auto
  note sNF = NF_instance[OF sNF[unfolded \<sigma>]]
  have edge2: "case mgu_class (icap_impl' U [] [] u)
           t of
     None \<Rightarrow> False
     | Some \<mu> \<Rightarrow>
         is_QNF d (mv_yvar s \<cdot> \<mu>)"
    unfolding icap_impl'_sound using mgu sNF uNF q by simp
  show ?thesis using edge edge2
    unfolding is_iedg_edge_dpp_def Let_def IEDG_def split U
    unfolding icap_impl_dpp_icap_mv[OF I, of d] by auto
qed

definition
  check_dep_graph_proc ::
    "('dpp, 'f::{show,linorder}, string) dpp_ops \<Rightarrow> 'dpp
    \<Rightarrow> ('a option \<times> ('f, string) rules) list \<Rightarrow> shows check"
where
  "check_dep_graph_proc I dpp dps = do {
    let c      = capRM_dpp I dpp;
    let rc     = reverse_capRM_dpp I dpp;
    let nfqsub = dpp_ops.NFQ_subset_NF_rules I dpp;
    let isnf   = dpp_ops.is_QNF I dpp;
    let ls     = map fst (dpp_ops.rules I dpp);
    let iedg   = is_iedg_edge_dpp I dpp;
    let P      = dpp_ops.pairs I dpp;
    check_subseteq P (concat (map snd dps))
      <+? (\<lambda>dp. ''Dependency Pair '' +#+ shows_rule dp +@+ '' is missing in decomposition''
        +#+ shows_nl);
    check_graph_decomp
      (shows \<circ> fst)
      (\<lambda>(((s, t), (_, ct, ict)), ((u, _), (cu, _, _))).
        Ground_Context.match ct u
        \<and> Ground_Context.match cu t
        \<and> ict u
      )
      (map (\<lambda>(real, Cs). (real \<noteq> None, (map (\<lambda>(s, t). ((s, t), (rc s, c t, iedg (s,t)))) Cs))) dps)
        <+? (\<lambda>s. ''our estimation (EDG*** + IEDG***) could not show that you have a valid decomposition ''
        +#+ ''due to the following reason'' +#+ shows_nl +@+ s)
  }"

lemma check_dep_graph_proc_sound:
  assumes I: "dpp_spec I"
    and check: "isOK (check_dep_graph_proc I d edps)" 
    and chain: "min_ichain (dpp_ops.dpp I d) s t \<sigma> "
  shows "\<exists>edp\<in>set edps. fst edp \<noteq> None
    \<and> (\<exists>i. min_ichain (dpp_ops.nfs I d, dpp_ops.minimal I d, set (snd edp), {}, set (dpp_ops.Q I d),
    set (dpp_ops.R I d), set (dpp_ops.Rw I d)) (shift s i) (shift t i) (shift \<sigma> i))"
proof -
  let ?c  = "capRM_dpp I d"
  let ?rc = "reverse_capRM_dpp I d"
  let ?Pb = "set (dpp_ops.pairs I d)"
  let ?Rb = "set (dpp_ops.rules I d)"
  let ?Q  = "set (dpp_ops.Q I d)"
  let ?R  = "set (dpp_ops.R I d)"
  let ?Rw = "set (dpp_ops.Rw I d)"
  let ?iedg   = "is_iedg_edge_dpp I d"
  let ?ls     = "map fst (dpp_ops.rules I d)"
  let ?qnil   = "dpp_ops.Q_empty I d"
  let ?nfs = "dpp_ops.nfs I d"
  let ?m = "dpp_ops.minimal I d"
  from check
    have subset: "set (dpp_ops.pairs I d) \<subseteq> set (concat (map snd edps))" (is "_ \<subseteq> ?dps")
    and scc: "isOK (check_graph_decomp (shows \<circ> fst)
    (\<lambda>(((s, t), (_, ct, ict)),((u, _), (cu, _, _))).
      Ground_Context.match ct u \<and> Ground_Context.match cu t
      \<and> ict u
    )
    (map (\<lambda>(real, Cs).
      (real \<noteq> None, (map (\<lambda>(s,t).
        ((s, t), (reverse_capRM_dpp I d s, capRM_dpp I d t,
                  ?iedg (s,t)))) Cs))) edps))"
    (is "isOK (check_graph_decomp ?sshow ?echeck (map ?conv edps))")
    unfolding check_dep_graph_proc_def
    by (auto simp: Let_def)
  let ?quads = "map ?conv edps"
  note sound = dpp_spec.dpp_spec_sound[OF I]
    dpp_spec.capRM_dpp[OF I] dpp_spec.reverse_capRM_dpp[OF I]
  from chain have inP: "\<And>i. (s i, t i) \<in> ?Pb"
    and betw: "\<And>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep ?nfs ?Q ?Rb)^*"
    and NFs: "\<And>i. s i \<cdot> \<sigma> i \<in> NF_terms ?Q" 
    and nfs: "\<And> i. NF_subst ?nfs (s i, t i) (\<sigma> i) ?Q"
    and SN: "\<And>i. ?m \<Longrightarrow> SN_on (qrstep ?nfs ?Q ?Rb) {t i \<cdot> \<sigma> i}"
    by (auto simp: ichain.simps sound minimal_cond_def)
  let ?p = "\<lambda>i. ((s i, t i), (?rc (s i), ?c (t i), ?iedg (s i,t i)))"
  let ?ps = "\<lambda>i. (s i \<cdot> \<sigma> i, t i \<cdot> \<sigma> i)"
  fix whatever
  {
    fix n
    from qrsteps_into_rsteps[OF betw[of n]]
      have steps: "(t n \<cdot> \<sigma> n, s (Suc n) \<cdot> \<sigma> (Suc n)) \<in> (rstep ?Rb)^* " by auto
    have "Ground_Context.match (Edg.cap ?Rb (t n)) (s (Suc n))
      \<and> Ground_Context.match (Edg.cap (?Rb^-1) (s (Suc n))) (t n)"
      using match_cap_sound[OF steps] match_cap_inv_sound[OF steps] by auto
    moreover
      from is_iedg_edge_dpp_sound[OF I NFs[of n] NFs nfs nfs SN betw]
      have "?iedg (s n, t n) (s (Suc n))" .
    ultimately have "Ground_Context.match (Edg.cap ?Rb (t n)) (s (Suc n))
      \<and> Ground_Context.match (Edg.cap (?Rb^-1) (s (Suc n))) (t n)
      \<and> ?iedg (s n, t n) (s (Suc n))"  by simp
  }
  note 1 = this
  have one: "\<And>n. ?echeck (?p n, ?p (Suc n))"
  proof -
    fix n
    from 1[of n]
    show "?echeck (?p n, ?p (Suc n))" unfolding sound by simp
  qed
  from inP subset have two: "path_nodes ?p \<subseteq> set (concat (map snd ?quads))"
    unfolding path_nodes_def
    by (auto split: prod.splits) fastforce
  from check_graph_decomp_sound[OF scc _ two, OF one]
  obtain q D where 
    suffix: "?p =\<dots> q" and 
    mem: "D \<in> set (map snd (filter fst ?quads))" and 
    ending: "path_nodes q \<subseteq> set D" by blast
  from mem obtain C S
    where C: "(S, C) \<in> set edps \<and> D = snd (?conv (S, C)) \<and> S \<noteq> None" by force
  from suffix obtain n where pq: "\<And>m. (?p (m + n) = q m)" unfolding suffix_def by blast
  {
    fix i
    have "(shift s n i, shift t n i) \<in> set C"
      using inP ending C pq[of i] unfolding path_nodes_def by force
  }
  note CP = this
  from chain CP
    have ichain: "min_ichain (?nfs,?m,set C, {}, ?Q, ?R, ?Rw) (shift s n) (shift t n) (shift \<sigma> n)"
    unfolding sound ichain.simps min_ichain.simps
    by (intro conjI allI, auto simp: minimal_cond_def)
  from ichain C
    have "\<exists>edp \<in> set edps. fst edp \<noteq> None
    \<and> min_ichain (?nfs,?m,set (snd edp), {}, ?Q, ?R, ?Rw) (shift s n) (shift t n) (shift \<sigma> n)"
    by force
  thus ?thesis by blast
qed

definition
  dep_graph_proc ::
    "('dpp, 'f::{show,linorder}, string) dpp_ops \<Rightarrow> 'dpp
    \<Rightarrow> ('a option \<times> ('f, string) rules) list \<Rightarrow> ('a \<times> 'dpp) list result"
where
  "dep_graph_proc I d dps \<equiv> check_return
    (check_dep_graph_proc I d dps)
    (map (\<lambda>aP. (the (fst aP), dpp_ops.intersect_pairs I d (snd aP)))
         (filter (\<lambda> aP. fst aP \<noteq> None) dps))"

lemma dep_graph_proc :
  assumes I: "dpp_spec I"
    and proc: "dep_graph_proc I d dps = return dpps"
    and rec: "\<And>a d'. (a, d') \<in> set dpps \<Longrightarrow> finite_dpp (dpp_ops.dpp I d')"
  shows "finite_dpp (dpp_ops.dpp I d)"
  unfolding finite_dpp_def
proof
  let ?P = "set (dpp_ops.P I d)"
  let ?Pw = "set (dpp_ops.Pw I d)"
  let ?Q = "set (dpp_ops.Q I d)"
  let ?R = "set (dpp_ops.R I d)"
  let ?Rw = "set (dpp_ops.Rw I d)"
  let ?nfs = "dpp_ops.nfs I d"
  let ?m = "dpp_ops.minimal I d"
  let ?Dpp = "(?nfs,?m,?P, ?Pw, ?Q, ?R, ?Rw)"
  let ?DPP = "dpp_ops.dpp I d"
  note sound = dpp_spec.dpp_spec_sound[OF I]
  assume "\<exists>s t \<sigma>. min_ichain ?DPP s t \<sigma>"
  from this obtain s t \<sigma> where michain: "min_ichain ?DPP s t \<sigma>" by auto
  hence ichain: "ichain ?DPP s t \<sigma>" and mc: "min_ichain ?Dpp s t \<sigma>"
    unfolding sound by auto
  from proc[unfolded dep_graph_proc_def]
    have check: "isOK (check_dep_graph_proc I d dps)"
    and dpps: "dpps = map (\<lambda>aP. (the (fst aP), dpp_ops.intersect_pairs I d (snd aP)))
                          (filter (\<lambda>aP. fst aP \<noteq> None) dps)"
    by auto
  from check_dep_graph_proc_sound[OF I check michain]
  obtain edp i where nearly: "edp \<in> set dps" "fst edp \<noteq> None"
    and ic: "min_ichain (?nfs,?m,set (snd edp), {}, ?Q, ?R, ?Rw) (shift s i) (shift t i) (shift \<sigma> i)"
    by auto
  hence ic: "ichain (?nfs,?m,set (snd edp), {}, ?Q, ?R, ?Rw) (shift s i) (shift t i) (shift \<sigma> i)" by simp
  from nearly have "(the (fst edp), dpp_ops.intersect_pairs I d (snd edp)) \<in> set dpps"
    unfolding dpps by auto
  from rec[OF this]
    have contra: "finite_dpp (?nfs,?m,?P \<inter> set (snd edp), ?Pw \<inter> set (snd (edp)), ?Q, ?R, ?Rw)" 
    using  dpp_spec.intersect_pairs_sound[OF I, of d "snd edp"]
    unfolding sound by auto
  from ichain_shift_merge[OF ic mc]
    have "min_ichain (?nfs,?m,?P \<inter> set (snd edp), ?Pw \<inter> set (snd (edp)), ?Q, ?R, ?Rw)
                     (shift s i) (shift t i) (shift \<sigma> i)" .
  with contra show False unfolding finite_dpp_def by blast
qed

end