(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory KBO_Impl
imports
  KBO
  Term_Impl
  Term_Order_Impl
begin

context weight_fun_prc
begin

lemma kbo_code: 
  "kbo s t = (if (multiset_of (vars_term_list' t) \<le> multiset_of (vars_term_list' s) \<and> weight t \<le> weight s)
      then if (weight t < weight s) then (True,True)
      else (case s of
      Var y \<Rightarrow> (False, (case t of Var x \<Rightarrow> True | Fun g ts \<Rightarrow> ts = [] \<and> least g)) 
    | Fun f ss \<Rightarrow> case t of Var x \<Rightarrow> (True,True)
                 | Fun g ts \<Rightarrow> let pf = prc (f,length ss); pg = prc (g,length ts) in if pf > pg then (True,True) else if pf \<ge> pg then 
              lex_ext_unbounded kbo ss ts else (False,False))
       else (False,False))" 
proof -
  {
    fix x y
    assume s: "s = Var x" and t: "t = Var y" and vars: "vars_term_ms t \<le> vars_term_ms s"
    from vars[unfolded s t] have "{#y#} \<le> {#x#}" by simp
    from mset_leD[OF this, of y, simplified] have "y = x" by (cases "x = y") auto
  }
  note main = this
  show ?thesis
    unfolding kbo.simps multiset_of_vars_term_list' Let_def 
    by (rule if_cong[OF refl if_cong[OF refl refl term.case_cong[OF refl _ refl]] refl], insert main, cases t, auto)
qed
end

abbreviation kbo where "kbo \<equiv> weight_fun_prc.kbo"
declare weight_fun_prc.weight.simps[code]
declare weight_fun_prc.kbo_code[code]

type_synonym 'f prec_weight_repr = "(('f \<times> nat) \<times> (nat \<times> nat))list \<times> nat"

fun shows_kbo_repr :: "('f::show)prec_weight_repr \<Rightarrow> shows"
where "shows_kbo_repr (prs,w0) = (''KBO with the following precedence and weight function'' +#+ shows_nl +@+ shows_map (\<lambda>((f,n),(pr,w)).
  ''precedence('' +#+ shows f +@+ ''['' +#+ shows n +@+ '']) = '' +#+ shows pr +@+ shows_nl
) prs +@+ shows_nl
+@+ ''precedence(_) = 0'' +#+ shows_nl +@+
  ''and the following weight'' +#+ shows_nl +@+
shows_map (\<lambda>((f,n),(pr,w)).
  ''weight('' +#+ shows f +@+ ''['' +#+ shows n +@+ '']) = '' +#+ shows w +@+ shows_nl
) prs +@+ shows_nl
+@+ ''weight(_) = '' +#+ shows (Suc w0) +@+ shows_nl +@+ shows ''w0 = '' +@+ shows w0 +@+ shows_nl
)"

definition
  prec_weight_repr_to_prec_weight :: "('f :: show) prec_weight_repr \<Rightarrow> shows check \<times> ('f \<times> nat \<Rightarrow> nat) \<times> ('f \<times> nat \<Rightarrow> nat) \<times> nat \<times> ('f \<Rightarrow> bool)" 
where
  "prec_weight_repr_to_prec_weight prw_w0 \<equiv>
    let 
      (prw, w0) = prw_w0;
      w_fun   = (\<lambda> fn. case map_of prw fn of Some (p,w) \<Rightarrow> w | None \<Rightarrow> Suc w0);
      p_fun   = (\<lambda> fn. case map_of prw fn of Some (p,w) \<Rightarrow> p | None \<Rightarrow> 0);
      fs      = map fst prw;
      p_max   = max_list (map p_fun fs);
      cw_okay = check_allm (\<lambda> fn. check (snd fn = 0 \<longrightarrow> w_fun fn \<ge> w0) (shows ''weight of constant '' +@+ shows (fst fn) +@+ shows '' must be at least w0'')) (map fst prw);
      adm     = check_allm (\<lambda> fn. check (snd fn = 1 \<longrightarrow> w_fun fn = 0 \<longrightarrow> p_fun fn = p_max) (shows ''unary symbol '' +@+ shows (fst fn) +@+ shows '' with weight 0 does not have maximal precedence'')) (map fst prw);
      cs      = map fst (filter (\<lambda> fn. snd fn = 0 \<and> w_fun fn = w0) fs);
      p_min   = min_list (map (\<lambda> c. p_fun (c,0)) cs);
      lcs     = filter (\<lambda> c. p_fun (c,0) = p_min) cs;
      least   = (\<lambda> c. c \<in> set lcs);
      ok      = (do {
        check (w0 > 0) (shows ''w0 must be larger than 0'');
        adm;
        cw_okay
      })
      in (ok, p_fun, w_fun, w0, least)"

lemma prec_weight_repr_to_prec_weight: 
  assumes ok: "prec_weight_repr_to_prec_weight prw_w0 = (OK,p,w,w0,least)"
  shows "admissible_weight_fun_prc w w0 p least"
proof -
  have id: "\<And> b. (b = OK) = isOK(b)" by auto
  obtain prw w0' where prw_w0: "prw_w0 = (prw,w0')" by force
  note ok = ok[unfolded prec_weight_repr_to_prec_weight_def prw_w0 Let_def split]
  from ok have w0': "w0' = w0" by simp
  note ok = ok[unfolded w0']
  obtain fs where fs: "fs = set (map fst prw)" by auto
  note ok = ok[simplified, unfolded id, simplified]
  from ok have w: "w = (\<lambda> fn. case map_of prw fn of Some (p,w) \<Rightarrow> w | None \<Rightarrow> Suc w0)" by simp
  from ok have p: "p = (\<lambda> fn. case map_of prw fn of Some (p,w) \<Rightarrow> p | None \<Rightarrow> 0)" by simp
  from ok have cw_okay: "\<And> fn. fn \<in> fs \<Longrightarrow> snd fn = 0 \<Longrightarrow> w fn \<ge> w0" unfolding w fs by auto
  obtain pmax where pmax: "pmax = max_list (map p (map fst prw))" by auto
  from ok have adm: "\<And> fn. fn \<in> fs \<Longrightarrow> snd fn = 1 \<Longrightarrow> w fn = 0 \<Longrightarrow> p fn = pmax"
    unfolding pmax p w fs by auto
  from ok have w0: "w0 > 0" by simp
  obtain cs where cs: "cs = map fst (filter (\<lambda> fn. snd fn = 0 \<and> w fn = w0) (map fst prw))" by auto
  obtain pmin where pmin: "pmin = min_list (map (\<lambda> c. p (c,0)) cs)" by auto
  obtain lcs where lcs: "lcs = filter (\<lambda> c. p (c,0) = pmin) cs" by auto
  have least: "least = (\<lambda> c. c \<in> set lcs)"
    using ok unfolding lcs pmin cs fs by auto
  {
    fix fn
    assume "fn \<notin> fs"
    hence fn: "\<And> e. (fn,e) \<notin> set prw" unfolding fs by force
    with map_of_SomeD[of prw fn] have l: "map_of prw fn = None" 
      by (cases "map_of prw fn", auto)
    hence "p fn = 0 \<and> w fn = Suc w0" unfolding p w by auto
  } note not_fs = this
  show ?thesis
  proof
    show "w0 > 0" by (rule w0)
  next    
    fix f
    show "w0 \<le> w (f,0)"
    proof (cases "(f,0) \<in> fs")
      case False
      from not_fs[OF this] show ?thesis by simp
    next
      case True
      from cw_okay[OF this] show ?thesis by simp
    qed
  next
    fix f g n
    assume wf: "w (f,1) = 0"
    show "p (g,n) \<le> p (f,1)" 
    proof (cases "(g,n) \<in> fs")
      case False
      from not_fs[OF this] show ?thesis by simp
    next
      case True
      from not_fs[of "(f,1)"] wf have f: "(f,1) \<in> fs" by auto
      from True have "p (g,n) \<in> set (map p (map fst prw))" unfolding fs by force
      from max_list[OF this] have "p (g,n) \<le> pmax" unfolding pmax .
      also have "... = p (f,1)" using adm[OF f _ wf] by simp
      finally show "p (g,n) \<le> p (f,1)" by simp
    qed
  next
    fix f
    show "least f = (w (f,0) = w0 \<and> (\<forall> g. w (g,0) = w0 \<longrightarrow> p (f,0) \<le> p (g,0)))" (is "_ = ?r")
    proof (cases "(f,0) \<in> fs \<and> w (f,0) = w0")
      case False
      hence "f \<notin> set cs" unfolding cs fs by force
      hence "least f = False" unfolding least lcs by simp
      also have "... = ?r"
      proof (cases "(f,0) \<in> fs")
        case True
        thus ?thesis using False by simp
      next
        case False
        thus ?thesis using not_fs[OF False] by simp
      qed
      finally show ?thesis .
    next
      case True
      let ?ps = "map (\<lambda> c. p (c,0)) cs"
      {
        fix c
        assume wc: "w (c,0) = w0"
        with not_fs[of "(c,0)"] have c: "(c,0) \<in> fs" by auto
        with wc have c: "c \<in> set cs" unfolding cs fs by force
        hence "p (c,0) \<in> set ?ps" by auto
        from pmin min_list[OF this] have "pmin \<le> p (c,0)" by simp
      } note pmin_c = this        
      from True have f: "f \<in> set cs" and wf: "w (f,0) = w0" unfolding cs fs by force+      
      show ?thesis 
      proof (cases "f \<in> set lcs")
        case True
        hence least: "least f = True" unfolding least by simp
        have ?r
        proof (rule conjI[OF wf], intro allI impI)
          fix g
          assume wg: "w (g,0) = w0"
          from pmin_c[OF this] have pg: "pmin \<le> p (g,0)" by simp
          from True[unfolded lcs] have pf: "p (f,0) = pmin" by simp
          from pf pg show "p (f,0) \<le> p (g,0)" by simp
        qed
        with least show ?thesis by simp
      next
        case False
        from False[unfolded lcs] f have "p (f,0) \<noteq> pmin" by force
        with pmin_c[OF wf] have pf: "p (f,0) > pmin" by auto
        from f have "?ps \<noteq> []" by auto
        from min_list_ex[OF this] obtain pm where pm: "pm \<in> set ?ps" 
          and pmin: "pm = pmin" unfolding pmin by auto
        hence pmin: "pmin \<in> set ?ps" by auto
        then obtain c where c: "c \<in> set cs" and pc: "p (c,0) = pmin" by auto
        from c[unfolded cs] have wc: "w (c,0) = w0" by auto
        from pf pc have p: "p (c,0) < p (f,0)" by simp
        from False p wc  show ?thesis unfolding least by auto
      qed
    qed
  qed
qed
                  
definition
  kbo_strict :: "('f \<times> nat \<Rightarrow> nat) \<Rightarrow> ('f \<times> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f \<Rightarrow> bool) \<Rightarrow> ('f :: show,'v :: show)rule \<Rightarrow> shows check"
where
  "kbo_strict pr w w0 least \<equiv> \<lambda>(s, t). check (fst (kbo w w0 pr least s t)) (''could not orient '' +#+ shows s +@+ '' >KBO '' +#+ shows t +@+ shows_nl)"

lemma kbo_strict[simp]: "isOK (kbo_strict pr w w0 least st) = fst (kbo w w0 pr least (fst st) (snd st))" unfolding kbo_strict_def by (cases st, auto)


definition
  kbo_nstrict :: "('f \<times> nat \<Rightarrow> nat) \<Rightarrow> ('f \<times> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f \<Rightarrow> bool) \<Rightarrow> ('f :: show,'v :: show)rule \<Rightarrow> shows check"
where
  "kbo_nstrict pr w w0 least \<equiv> \<lambda>(s, t). check (snd (kbo w w0 pr least s t)) (''could not orient '' +#+ shows s +@+ '' >=KBO '' +#+ shows t +@+ shows_nl)"

lemma kbo_nstrict[simp]: "isOK (kbo_nstrict pr w w0 least st) = snd (kbo w w0 pr least (fst st) (snd st))" unfolding kbo_nstrict_def by (cases st, auto)

fun create_KBO_redtriple :: "(('f :: show)prec_weight_repr \<Rightarrow> 'g prec_weight_repr) \<Rightarrow> 'f prec_weight_repr \<Rightarrow> ('g :: show,'v :: show)redtriple"
where "create_KBO_redtriple f_to_g pr = (let (ch,p,w,w0,least) = prec_weight_repr_to_prec_weight (f_to_g pr);
   ns = kbo_nstrict p w w0 least;
   s =  kbo_strict p w w0 least
    in
  \<lparr> redtriple.valid = ch, s = s, ns = ns, nst = ns, af = full_af, mono = (\<lambda> _. OK), desc =shows_kbo_repr pr, cpx = no_complexity\<rparr>)"

interpretation KBO_redpair: generic_redtriple_impl "create_KBO_redtriple (f_to_g :: ('f :: show)prec_weight_repr \<Rightarrow> ('g :: show)prec_weight_repr)"
proof
  fix "prw" :: "('f :: show)prec_weight_repr"
   and s_list ns_list nst_list :: "('g :: show,'v :: show)rule list"
  let ?rp = "create_KBO_redtriple f_to_g prw :: ('g,'v)redtriple"
  let ?af = "redtriple.af ?rp :: ('g af)"
  let ?pr = "prec_weight_repr_to_prec_weight (f_to_g prw)"
  let ?cpx = "redtriple.cpx ?rp"
  let ?all = "s_list @ ns_list @ nst_list"
  obtain ch "pr" w w0 least where id: "?pr = (ch,pr,w,w0,least)" by (cases ?pr, force)
  assume valid: "isOK(redtriple.valid ?rp)"
   and stri: "isOK(check_allm (redtriple.s ?rp) s_list)"
   and nstri: "isOK (check_allm (redtriple.ns ?rp) ns_list)"
   and nstri_top: "isOK (check_allm (redtriple.nst ?rp) nst_list)"
  have af: "?af = full_af" by (simp add: id Let_def)
  from valid have "isOK(ch)" by (simp add: id Let_def)
  hence ch: "ch = OK" by (cases ch, auto)
  have cpx: "?cpx = no_complexity" by (simp add: id Let_def)
  note id = id[unfolded ch cpx]
  interpret admissible_weight_fun_prc w w0 "pr" least 
    by (rule prec_weight_repr_to_prec_weight[OF id])
  from kbo_redpair 
  interpret mono_ce_af_redtriple_order kbo_S kbo_NS kbo_NS full_af .
  show "\<exists> S NS NST. cpx_ce_af_redtriple_order S NS NST ?af ?cpx \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> set nst_list \<subseteq> NST \<and> (isOK (redtriple.mono ?rp ?all) \<longrightarrow> mono_ce_af_redtriple_order S NS NST ?af \<and> ctxt_closed S)"
    unfolding af cpx
  proof (intro exI conjI impI)
    show "cpx_ce_af_redtriple_order kbo_S kbo_NS kbo_NS full_af no_complexity" ..
    show "mono_ce_af_redtriple_order kbo_S kbo_NS kbo_NS full_af" ..
    show "ctxt_closed kbo_S" by (rule kbo_S_mono)
    show "set s_list \<subseteq> kbo_S" using stri by (force simp: Let_def id)
    show "set ns_list \<subseteq> kbo_NS" using nstri by (force simp: Let_def id)
    show "set nst_list \<subseteq> kbo_NS" using nstri_top by (force simp: Let_def id)
  qed
qed

end