(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Reduction_Pair_Implementations
imports
  Linear_Poly_Carrier
  Argument_Filter
  Poly_Order
  Efficient_RPO
  SCNP_Impl
  List_Order_Implementations
  KBO_Impl
begin

datatype 'f redtriple_impl = 
    Int_carrier "('f, int) lpoly_interL" 
  | Int_nl_carrier "('f, int) poly_inter_list" 
  | Rat_carrier "('f, rat) lpoly_interL"
  | Rat_nl_carrier rat "('f, rat) poly_inter_list"
  | Arctic_carrier "('f, arctic) lpoly_interL" 
  | Arctic_rat_carrier "('f, arctic_rat) lpoly_interL" 
  | Int_mat_carrier nat nat "('f, int mat) lpoly_interL"
  | Rat_mat_carrier nat nat "('f, rat mat) lpoly_interL"
  | Arc_mat_carrier nat "('f, arctic mat) lpoly_interL"
  | Arc_rat_mat_carrier nat "('f, arctic_rat mat) lpoly_interL"
  | RPO "'f status_prec_repr" "'f afs_list"
  | KBO "'f prec_weight_repr" "'f afs_list"

definition faulty_redtriple :: "string \<Rightarrow> ('f :: show, 'v :: show)redtriple"
where "faulty_redtriple s \<equiv> \<lparr>redtriple.valid = error (shows_string s), s = \<lambda> _. OK, ns = \<lambda> _. OK, nst = \<lambda> _. OK, af = full_af, mono = \<lambda> _. OK, desc = id, cpx = no_complexity \<rparr>"

lemma faulty_redtriple: "generic_redtriple_impl faulty_redtriple"
  by (unfold_locales, unfold faulty_redtriple_def, auto)

fun filter_prec_weight_repr :: "'f afs \<Rightarrow> 'f prec_weight_repr \<Rightarrow> 'f filtered prec_weight_repr"
where "filter_prec_weight_repr \<pi> (prw,w0) = (
   let fprw = filter (\<lambda> ((f,n),e). case \<pi> f n of Collapse _ \<Rightarrow> False | AFList _ \<Rightarrow> True) prw;
       mprw = map (\<lambda> ((f,n),e). ((FPair f n, case \<pi> f n of AFList l \<Rightarrow> length l | Collapse _ \<Rightarrow> 0),e)) fprw
    in (mprw,w0))"

fun prec_repr_to_pr :: "'f status_prec_repr \<Rightarrow> ('f filtered \<Rightarrow> nat \<Rightarrow> nat)" where
  "prec_repr_to_pr prs (FPair f a) _ = (case map_of prs (f, a) of None \<Rightarrow> 0 | Some x \<Rightarrow> fst x)" 

fun prec_repr_to_status :: "'f status_prec_repr \<Rightarrow> ('f filtered \<Rightarrow> nat \<Rightarrow> order_tag)" where
  "prec_repr_to_status prs (FPair f a) _ = (case map_of prs (f, a) of None \<Rightarrow> Lex | Some x \<Rightarrow> snd x)" 

fun get_redtriple :: "('f :: show) redtriple_impl \<Rightarrow> ('f,'v :: show)redtriple"
where 
  "get_redtriple (Int_carrier I) = int_redtriple I"
| "get_redtriple (Int_nl_carrier I) = create_nlpoly_redtriple (1 :: int) (op >) True True I"
| "get_redtriple (Rat_carrier I) = rat_redtriple I"
| "get_redtriple (Rat_nl_carrier d I) = (if d > 0 then create_nlpoly_redtriple d (rat_gt d) (1 \<le> d) False I else faulty_redtriple ''delta must be larger than zero'')"
| "get_redtriple (Arctic_carrier I) = arctic_redtriple I"
| "get_redtriple (Arctic_rat_carrier I) = arctic_rat_redtriple I"
| "get_redtriple (Int_mat_carrier n sd I) = int_mat_redtriple n sd I"
| "get_redtriple (Rat_mat_carrier n sd I) = rat_mat_redtriple n sd I"
| "get_redtriple (Arc_mat_carrier n I) = arctic_mat_redtriple n I"
| "get_redtriple (Arc_rat_mat_carrier n I) = arctic_rat_mat_redtriple n I"
| "get_redtriple (RPO prec\<tau> pi) = af_redtriple pi (create_RPO_redtriple (\<lambda> pr. (prec_repr_to_pr pr, prec_repr_to_status pr)) prec\<tau>)"
| "get_redtriple (KBO precw pi) = af_redtriple pi (create_KBO_redtriple (filter_prec_weight_repr (afs_list_to_afs pi)) precw)"

lemma get_redtriple: "generic_redtriple_impl get_redtriple"
proof
  fix rp :: "('f :: show)redtriple_impl" and s_list ns_list nst_list:: "('f,'v::show)rules"
  let ?rp = "get_redtriple rp :: ('f,'v)redtriple"
  let ?all = "s_list @ ns_list @ nst_list"
  assume v: "isOK(redtriple.valid ?rp)" and S: "isOK(check_allm (redtriple.s ?rp) s_list)" and NS: "isOK(check_allm (redtriple.ns ?rp) ns_list)"
    and NST: "isOK(check_allm (redtriple.nst ?rp) nst_list)"
  show "\<exists> S NS NST. cpx_ce_af_redtriple_order S NS NST (redtriple.af ?rp) (redtriple.cpx ?rp) \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> set nst_list \<subseteq> NST \<and> (isOK(redtriple.mono ?rp ?all) \<longrightarrow> mono_ce_af_redtriple_order S NS NST (redtriple.af ?rp) \<and> ctxt_closed S)"
  proof (cases rp)
    case (Int_carrier I)
    hence id: "?rp = (int_redtriple :: ('f,int) lpoly_interL \<Rightarrow> ('f,'v)redtriple) I" (is "_ = ?I I") by simp    
    have rp: "generic_redtriple_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]])
  next
    case (Int_nl_carrier I)
    hence id: "?rp = (create_nlpoly_redtriple (1 :: int) (op >) True True :: ('f,int)poly_inter_list \<Rightarrow> ('f,'v)redtriple) I" (is "_ = ?I I") by simp
    have "SN_Orders.poly_order_carrier (1 :: int) (op >) True True" ..
    from poly_order_carrier_to_generic_redtriple_impl[OF this]
    have rp: "generic_redtriple_impl ?I" .
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]]) 
  next
    case (Rat_carrier I)
    hence id: "?rp = (rat_redtriple :: ('f,rat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) I" (is "_ = ?I I") by simp    
    have rp: "generic_redtriple_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]]) 
  next
    case (Rat_nl_carrier d I)
    show ?thesis 
    proof (cases "d > 0")
      case True
      with Rat_nl_carrier have id: "?rp = (create_nlpoly_redtriple d (rat_gt d) (1 \<le> d) False :: ('f,rat)poly_inter_list \<Rightarrow> ('f,'v)redtriple) I" (is "_ = ?I I") by simp
      from poly_order_carrier_to_generic_redtriple_impl[OF rat_poly[OF True, of d]]
      have rp: "generic_redtriple_impl ?I" by simp
      show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]])
    next       
      case False with Rat_nl_carrier v have False by (simp add: faulty_redtriple_def) 
      thus ?thesis ..
    qed
  next
    case (Arctic_carrier I)
    hence id: "?rp = (arctic_redtriple :: ('f,arctic) lpoly_interL \<Rightarrow> ('f,'v)redtriple) I" (is "_ = ?I I") by simp    
    have rp: "generic_redtriple_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]]) 
  next
    case (Arctic_rat_carrier I)
    hence id: "?rp = (arctic_rat_redtriple :: ('f,arctic_rat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) I" (is "_ = ?I I") by simp    
    have rp: "generic_redtriple_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]]) 
  next
    case (Int_mat_carrier n sd I)
    hence id: "?rp = (int_mat_redtriple :: nat \<Rightarrow> nat \<Rightarrow> ('f,int mat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) n sd I" (is "_ = ?I I") by simp    
    have rp: "generic_redtriple_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]]) 
  next
    case (Rat_mat_carrier n sd I)
    hence id: "?rp = (rat_mat_redtriple :: nat \<Rightarrow> nat \<Rightarrow> ('f,rat mat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) n sd I" (is "_ = ?I I") by simp    
    have rp: "generic_redtriple_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]]) 
  next
    case (Arc_mat_carrier n I)
    hence id: "?rp = (arctic_mat_redtriple :: nat \<Rightarrow> ('f,arctic mat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) n I" (is "_ = ?I I") by simp    
    have rp: "generic_redtriple_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]]) 
  next
    case (Arc_rat_mat_carrier n I)
    hence id: "?rp = (arctic_rat_mat_redtriple :: nat \<Rightarrow> ('f,arctic_rat mat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) n I" (is "_ = ?I I") by simp    
    have rp: "generic_redtriple_impl ?I" ..
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?I, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]]) 
  next
    case (RPO prec\<tau> pi)
    let ?one = "(\<lambda> I. create_RPO_redtriple (\<lambda>pr. (prec_repr_to_pr pr, prec_repr_to_status pr)) I) :: 'f status_prec_repr \<Rightarrow> ('f filtered,'v)redtriple"
    have rp: "generic_redtriple_impl ?one" ..
    let ?two = "\<lambda> I. af_redtriple pi (?one I)"
    from af_redtriple[OF rp] 
    have rp: "generic_redtriple_impl ?two" .
    from RPO have id: "?rp = ?two prec\<tau>" by simp
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?two, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]])
  next
    case (KBO precw pi)
    let ?one = "(\<lambda> I. create_KBO_redtriple (filter_prec_weight_repr (afs_list_to_afs pi)) I) :: 'f prec_weight_repr \<Rightarrow> ('f filtered,'v)redtriple"
    have rp: "generic_redtriple_impl ?one" ..
    let ?two = "\<lambda> I. af_redtriple pi (?one I)"
    from af_redtriple[OF rp] 
    have rp: "generic_redtriple_impl ?two" .
    from KBO have id: "?rp = ?two precw" by simp
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?two, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]])
  qed 
qed

datatype 'f root_redtriple_impl = SCNP list_order_type "'f scnp_af" "'f redtriple_impl"

fun get_root_redtriple :: "('f :: show) root_redtriple_impl \<Rightarrow> ('f,'v :: show)root_redtriple"
where 
  "get_root_redtriple (SCNP type af rp) = generate_scnp_rp (list_ext (scnp_arity af) type) (list_ext_name type) af get_redtriple rp"

lemma get_root_redtriple: "generic_root_redtriple_impl get_root_redtriple"
proof
  fix rrp :: "('f :: show) root_redtriple_impl" and s_list ns_list nst_list :: "('f, 'v :: show)rule list"
  let ?r = "get_root_redtriple rrp :: ('f,'v)root_redtriple"  
  assume ass: "isOK(root_redtriple.valid ?r)" 
         "isOK(check_allm (root_redtriple.s ?r) s_list)"
         "isOK(check_allm (root_redtriple.ns ?r) ns_list)" 
         "isOK(check_allm (root_redtriple.nst ?r) nst_list)" 
  show "\<exists> s ns nst. ce_af_root_redtriple_order s ns nst (root_redtriple.af ?r) (root_redtriple.aft ?r) \<and> set s_list \<subseteq> s \<and> set ns_list \<subseteq> ns \<and> set nst_list \<subseteq> nst"
  proof (cases rrp)
    case (SCNP t af rp)
    show ?thesis unfolding SCNP get_root_redtriple.simps
      by (rule generic_root_redtriple_impl.generate_root_redtriple
      [OF generate_scnp_rp[OF get_redtriple list_ext], OF ass[unfolded SCNP get_root_redtriple.simps]])
  qed
qed 
end